/****************************************************************************
 *
 * main.c
 *
 * This file is part of DDM
 *
 * Copyright (C) 2013, 2016 Moreno Marzolla
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 ****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#ifdef _OPENMP
#include <omp.h>
#endif
#include <unistd.h>
#include "timing.h"
#include "interval.h"
#include "brute_force_matching.h"
#include "parallel_sort_matching.h"
#include "interval_tree_matching.h"
#include "cflags.h"

extern const char* CFLAGS;

const double Lmax = 1.0e6; /* total length */

/**
 * allocates and returns a vector of n random intervals of length
 * |l|. The lower bound of each interval is randomly placed in [0,
 * Lmax-l). The caller owns the returned pointer, and is therefore
 * responsible for freeing the allocated memory.
 */
struct interval* mk_random_interval( size_t n, double l )
{
    struct interval* result = (struct interval*)malloc( n * sizeof(result[0]) );
    size_t i;
    assert( l > 0 );
    for ( i=0; i<n; ++i ) {
	result[i].lower = (Lmax-l)*random() / ((double)RAND_MAX + 1);
	result[i].upper = result[i].lower + l;
        assert( result[i].lower < result[i].upper );
        result[i].payload = NULL;
    }
    return result;
}

int main( int argc, char* argv[] )
{
    char fname[256];
    const char* algo;
    int opt;
    size_t tot_extents = 100000;
    size_t d = 1; /* number of dimensions, must be 1 */
    size_t u; /* updates */
    size_t s; /* subscriptions */
    double alpha = 0.01; /* overlap ratio */
    size_t (*ddm_matching)( const struct interval* subscriptions, size_t n,
                            const struct interval* updates, size_t m );

    algo = "brute_force";
    ddm_matching = brute_force_matching;

    while ((opt = getopt(argc, argv, "hs:n:d:a:")) != -1) {
        switch (opt) {
        case 's':
            if ( 0 == strcmp(optarg, "interval_tree" )) {
                ddm_matching = interval_tree_matching;
                algo = "interval_tree";
            } else if (0 == strcmp(optarg, "parallel_sort")) {
                ddm_matching = parallel_sort_matching;
                algo = "parallel_sort";
            } else if ( 0 == strcmp(optarg, "brute_force")) {
                ddm_matching = brute_force_matching;
                algo = "brute_force";
            } else {
                fprintf(stderr, "\n%s error: invalid algorithm %s\n\nValid options: interval_tree, parallel_sort, brute_force (default)\n\n", argv[0], optarg);
                exit(EXIT_FAILURE);
            }
            break;
        case 'n':
            tot_extents = atol(optarg);
            break;
        case 'd':
            d = atoi(optarg);
            if ( d != 1 ) {
                fprintf(stderr, "This program only supports 1 dimension\n");
                exit(EXIT_FAILURE);
            }
            break;
        case 'a':
            alpha = atof(optarg);
            if (alpha <= 0.0) {
                fprintf(stderr, "Invalid alpha (%s): must be strictly positive\n", optarg);
                exit(EXIT_FAILURE);
            }
            break;
        default:
            fprintf(stderr, "Usage: %s [-s algo] [-n extents] [-d dims] [-a alpha]\n\nValid values for algo: brute_force (default), interval_tree, parallel_sort\n", argv[0]);
            exit(EXIT_FAILURE);
        }
    }

    printf("# %s compiled with:\n", argv[0]);
    printf("#   CFLAGS=%s\n", CFLAGS);
    printf("#   CXXFLAGS=%s\n", CXXFLAGS);
    printf("#   LDFLAGS=%s\n", LDFLAGS);
    printf("#   LDLIBS=%s\n", LDLIBS);
#ifdef _OPENMP
    printf("# OPENMP enabled with max %d threads\n", omp_get_max_threads());
#endif
    printf("# Algorithm ........... %s\n", algo);
    printf("# Number of extents.... %zu\n", tot_extents);
    printf("# Number of dimensions. %zu\n", d);
    printf("# alpha................ %f\n", alpha);
    srandom(clock());

    u = s = tot_extents / 2;
    double l = alpha * Lmax / (double)(u+s);
    assert( l > 0 );
    struct interval* upds = mk_random_interval( u, l );
    struct interval* subs = mk_random_interval( s, l );
    struct timer timing;

    timing_init( &timing );

    timing_start( &timing );
    size_t nmatches = ddm_matching( subs, s, upds, u );
    timing_stop( &timing );

    /*
    size_t nmatches_bf = brute_force_matching( subs, s, upds, u );
    size_t nmatches_it = interval_tree_matching( subs, s, upds, u );
    */

    double total_time = timing_get_average( &timing );

    /* write results */
    snprintf(fname, sizeof(fname), "%s.txt", algo);
    FILE* fout = fopen(fname, "a");
    if ( fout == NULL ) {
	fprintf(stderr, "Error appending to file %s\n", fname);
	exit(-1);
    }

    fprintf(fout, "%f\n", total_time);
    fclose(fout);

    printf("%zu matches %f\n", nmatches, total_time );

    return 0;
}
