/****************************************************************************
 *
 * parallel_sort_matching.cc
 *
 * Parallel sort-based matching
 *
 * Copyright (C) 2016 Moreno Marzolla
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 ****************************************************************************/

#include <vector>
#include <string>
#include <set>
#include <cassert>
#include <iostream>

#ifdef _OPENMP

#include <parallel/algorithm>
using __gnu_parallel::sort;
#include <omp.h>

#else

#include <algorithm>
using std::sort;
int omp_get_max_threads() { return 1; }
int omp_get_num_threads() { return 1; }
int omp_get_thread_num() { return 0; }

#endif // _OPENMP

#include "interval.h"
#include "parallel_sort_matching.h"
#include <time.h>

/******************************************************************************
 * Simple timing class. You create an instance of this class at
 * the beginning of the block you want to time. When this object
 * goes out of scope, it prints the elapsed time (in seconds).
 ******************************************************************************/
class omp_timer {
protected:
    const std::string name;
    struct timespec t_start;
    
    double elapsed( ) const {
        struct timespec t_stop;
        clock_gettime(CLOCK_MONOTONIC, &t_stop);
        return t_stop.tv_sec + (double)(t_stop.tv_nsec) / 1.0e9 - 
            (t_start.tv_sec + (double)(t_start.tv_nsec) / 1.0e9);
    }

public:
    omp_timer( const std::string& n ) : name(n) {
        clock_gettime(CLOCK_MONOTONIC, &t_start);
    }
    ~omp_timer( ) {
        struct timespec t_stop;
        clock_gettime(CLOCK_MONOTONIC, &t_stop);
        std::cout << "Timer[" << name << "] " << elapsed() << std::endl;
    }
    void tic( const std::string& msg ) const {
        std::cout << "Timer[" << name << "," << msg << "] " << elapsed() << std::endl;
    }
};

/******************************************************************************
 * Endpoint stored in the endpoint arrays. Each endpoint represents a
 * single scalar value v, that can either correspond to a lower or
 * upper bound of a subscription or update extent.
 ******************************************************************************/
struct endpoint {
        
    enum ep_extreme { LOWER, UPPER, UNDEF }; 
    enum ep_type { SUBSCRIPTION, UPDATE };
        
    size_t id;  // ID of the interval this endpoint belongs to; the id is the index (position) of the interval in the sub[] or upd[] arrays passed as parameter to function parallel_sort_matching()
    double v;   // value of this endpoint
    ep_extreme e; // whether this is a lower or upper endpoint of the interval
    ep_type t;  // whether this endpoint belongs to a subscription or update interval
        
    // Default constructor
    endpoint() :
        id(0), v(0.0), e(UNDEF), t(SUBSCRIPTION)
    { };
        
    // Constructor
    endpoint(size_t ep_id, double ep_v, ep_extreme ep_e, ep_type ep_t) :
        id( ep_id ), v( ep_v ), e( ep_e ), t( ep_t ) 
    { };
        
#if 0
    // Copy costructor
    endpoint( const endpoint& other) :
        id( other.id ),
        v( other.v ),
        e( other.e ),
        t( other.t )
    { };
#endif

    /* Endpoints are sorted according to their value. Special care
       must be taken for endpoints having exactly the same value.
       Since we are dealing with closed intervals, a lower
       endpoint must always precede an upper endpoint with the
       same value. If this does not happen, an overlap is missed.
    */
    bool operator < (const endpoint& other) const
    { 
        assert( UNDEF != e );
        assert( UNDEF != other.e );
        return (v < other.v || (v == other.v && e == LOWER && other.e == UPPER));
    }
};

#if defined(EPSET_VECTOR_BOOL)
/******************************************************************************
 * A epset (endpoint set) is a subset of the n integers {0, 1,
 * ... n-1}. epsets support insertion and removal of a single
 * endpoint, and union/difference with another epset. Initially you
 * create a epset specifying the value n; the set is initially empty
 * (it does not contain any integer). We provide several
 * implementations of epset. This one is based on a bit vector
 * implemented using the std::vector<bool> container.
 ******************************************************************************/
class epset { 
protected: 
    size_t sz;      // max n. of elements in the set 
    size_t count;   // number of elements in the set 
    std::vector<bool> data;
    bool dirty;     // true if count must be recomputed 
public:
    epset( size_t n ) : sz(n), count(0), data(n, false), dirty(false) { }
    epset& merge( const epset& other ) {
        assert(sz == other.sz);
        std::transform(data.begin(), data.end(),
                       other.data.begin(), data.begin(),
                       std::logical_or<bool>());
        dirty = true;
        return *this;
    }
    epset& subtract( const epset& other ) {
        assert(sz == other.sz);
        for (size_t i=0; i<sz; i++) {
            if (data[i] && other.data[i]) {
                data[i] = false;
                if ( !dirty ) count += data[i];
            }
        }
        return *this;
    }
    epset& clear( ) {
        std::fill(data.begin(), data.end(), false);
        count = 0;
        dirty = false;
        return *this;
    }
    // Insert v into the set (does nothing if already present)
    epset& insert( size_t v ) {
        assert( v >= 0 && v < sz );
        if ( false == data[v] ) {
            data[v] = true;
            if ( !dirty ) count++;
        }
        return *this;
    }
    // Remove v from the set (does nothing if not present)
    epset& remove( size_t v ) {
        assert( v >= 0 && v < sz );
        if ( data[v] ) {
            data[v] = false;
            if (!dirty) count--;
        }
        return *this;
    }
    // number of elements
    size_t size( void ) {
        if ( dirty ) 
            update_count();
        return count;
    }
    bool find( size_t v ) const {
        return data[v];
    }
    bool empty( ) {
        return (0 == size());
    }
protected:
    void update_count( void ) 
    {
        count = 0;
        for (size_t i=0; i<sz; i++) {
            count += data[i];
        }
        dirty = false;
    }
};
#elif defined(EPSET_SET)
/******************************************************************************
 * epset implemented using std::set (ordered set, generally realized
 * using RB Trees)
 ******************************************************************************/
class epset { 
protected: 
    size_t sz;
    std::set<int> data;
public:
    typedef std::set<int>::iterator iterator;
    typedef std::set<int>::const_iterator const_iterator;

    epset( size_t n ) : sz(n), data() { };
    epset& merge( const epset& other ) {
        data.insert( other.data.begin(), other.data.end() );
        return *this;
    }
    epset& subtract( const epset& other ) {
        std::set<int> result;
        std::set_difference(data.begin(), data.end(),
                            other.data.begin(), other.data.end(),
                            std::inserter(result, result.begin()));
        data = result;
        return *this;
    }
    epset& clear( ) {
        data.clear();
        return *this;
    }
    // Insert v into the epset (does nothing if already present)
    epset& insert( size_t v ) {
        assert( v >= 0 && v < sz );
        assert( data.find(v) == data.end() );
        data.insert(v);
        assert( data.find(v) != data.end() );
        return *this;
    }
    // Remove v from the epset (does nothing if not present)
    epset& remove( size_t v ) {
        assert( v >= 0 && v < sz );
        assert( data.find(v) != data.end() );
        data.erase(v);
        assert( data.find(v) == data.end() );
        return *this;
    }
    // number of elements
    size_t size( void ) {
        return data.size();
    }
    bool find( size_t v ) const {
        return (data.find(v) != data.end());
    }
    bool empty( void ) const {
        return (0 == data.size()); 
    }
};
#elif defined(EPSET_SET_UNORDERED)
/******************************************************************************
 * Epset implementation based on an unordered set; requires support
 * for C++11, as the std::unordered_set<> container is only available
 * in C++11
 ******************************************************************************/
#if __cplusplus < 201103L
#error You need C++11 for the unordered_set container
#else
#include <unordered_set>

class epset { 
protected: 
    size_t sz;
    std::unordered_set<int> data;
public:
    typedef std::unordered_set<int>::iterator iterator;
    typedef std::unordered_set<int>::const_iterator const_iterator;

    epset( size_t n ) : sz(n), data() { };
    epset& merge( const epset& other ) {
        data.insert( other.data.begin(), other.data.end() );
        return *this;
    }
    epset& subtract( const epset& other ) {
        // TODO: optimization if sets have different size (outer loop
        // should be performed on the smaller one).
        for (const_iterator it = other.data.begin(); it != other.data.end(); ++it ) {
            data.erase(*it);
        }
        return *this;
    }
    epset& clear( ) {
        data.clear();
        return *this;
    }
    // Insert v into the epset (does nothing if already present)
    epset& insert( size_t v ) {
        assert( v >= 0 && v < sz );
        assert( data.find(v) == data.end() );
        data.insert(v);
        assert( data.find(v) != data.end() );
        return *this;
    }
    // Remove v from the epset (does nothing if not present)
    epset& remove( size_t v ) {
        assert( v >= 0 && v < sz );
        assert( data.find(v) != data.end() );
        data.erase(v);
        assert( data.find(v) == data.end() );
        return *this;
    }
    // number of elements
    size_t size( void ) {
        return data.size();
    }
    bool find( size_t v ) const {
        return (data.find(v) != data.end());
    }
    bool empty( void ) const {
        return (0 == data.size());
    }
};
#endif
#elif defined(EPSET_RAW)
/******************************************************************************
 * epset implemented using a buffer of integers. Similar to
 * vector<bool>, but should have lower overhead.
 ******************************************************************************/
class epset { 
protected: 
    size_t sz;      /* max n. of elements in the set */
    size_t buflen;  /* Length of data buffer */
    size_t count;   /* number of values in the set */
    unsigned int *data;
public:
    epset( size_t n ) : 
        sz(n),
        buflen((n + sizeof(*data) - 1)/sizeof(*data)), 
        count(0), 
        data( new unsigned int[buflen] ) 
    { 
        clear();
    }
    // copy costructor
    epset( const epset& other ) :
        sz(other.sz),
        buflen(other.buflen),
        count(other.count),
        data( new unsigned int[buflen] ) 
    {
        // memcpy( data, other.data, buflen*sizeof(*data) );
        std::copy(other.data, other.data + other.buflen, data);
    }
    virtual ~epset( ) {
        delete[] data;
    }
    epset& merge( const epset& other ) {
        assert(sz == other.sz);
        std::transform(data, data+buflen,
                       other.data, 
                       data,
                       std::bit_or<unsigned int>());
        update_count();
        return *this;
    }
    epset& subtract( const epset& other ) {
        assert(sz == other.sz);
        count = 0;
        for (size_t i=0; i<buflen; i++) {
            data[i] = data[i] & ~other.data[i];
            count += __builtin_popcount(data[i]);
        }
        return *this;
    }
    epset& clear( ) {
        std::fill(data, data+buflen, 0u);
        count = 0;
        return *this;
    }
    // Insert v into the epset (does nothing if already present)
    epset& insert( size_t v ) {
        assert( v >= 0 && v < sz );
        size_t idx = v / sizeof(*data);
        assert( idx < buflen );
        count -= __builtin_popcount(data[idx]);
        unsigned int mask = (1u) << (v % sizeof(*data));
        data[idx] |= mask;
        count += __builtin_popcount(data[idx]);
        return *this;
    }
    // Remove v from the epset (does nothing if not present)
    epset& remove( size_t v ) {
        assert( v >= 0 && v < sz );
        size_t idx = v / sizeof(*data);
        assert( idx < buflen );
        count -= __builtin_popcount(data[idx]);
        unsigned int mask = (1u) << (v % sizeof(*data));
        data[idx] &= ~mask;
        count += __builtin_popcount(data[idx]);
        return *this;
    }
    bool find( size_t v ) const {
        assert( v >= 0 && v < sz );
        size_t idx = v / sizeof(*data);
        assert( idx < buflen );
        unsigned int mask = (1u) << (v % sizeof(*data));
        return (data[idx] & mask);
    }
    // number of elements
    size_t size( void ) const {
        return count;
    }
    bool empty( void ) const {
        return (0 == size());
    }
protected:
    void update_count( void ) 
    {
        count = 0;
        for (size_t i=0; i<buflen; i++) {
            count += __builtin_popcount(data[i]);
        }
    }
};
#else
#error Undefined set implementation
#endif

#if 1
#define BEGIN_TIMER(n) { omp_timer my_timer(n);
#define END_TIMER() }
#else
#define BEGIN_TIMER(n)
#define END_TIMER()
#endif

extern "C" 
size_t parallel_sort_matching( const struct interval* sub, size_t n,
                               const struct interval* upd, size_t m )
{
    omp_timer sbm_timer("Parallel SBM");
    const size_t n_endpoints = 2*(n+m);
    std::vector<endpoint> endpoints(n_endpoints);
    size_t nmatches = 0;

    // Push subscription intervals into the endpoints array (in parallel)
    BEGIN_TIMER("Fill");
#pragma omp parallel
    {
#pragma omp for 
        for (size_t i=0; i<n; i++) {
            assert(sub[i].lower < sub[i].upper);
            endpoints[i] = endpoint(i, sub[i].lower, endpoint::LOWER, endpoint::SUBSCRIPTION);
            endpoints[i+n] = endpoint(i, sub[i].upper, endpoint::UPPER, endpoint::SUBSCRIPTION);
        }
            
#pragma omp for 
        for (size_t i=0; i<m; i++) {
            assert(upd[i].lower < upd[i].upper);
            endpoints[2*n+i] = endpoint(i, upd[i].lower, endpoint::LOWER, endpoint::UPDATE);
            endpoints[2*n+m+i] = endpoint(i, upd[i].upper, endpoint::UPPER, endpoint::UPDATE);
        }
    }
    END_TIMER();

    // Sort endpoints (in parallel)
    BEGIN_TIMER("Sort");
    sort(endpoints.begin(), endpoints.end());
    END_TIMER();

    const int maxproc = omp_get_max_threads();

    std::vector< epset > sub_insert(maxproc, epset(n));
    std::vector< epset > sub_delete(maxproc, epset(n));
    std::vector< epset > upd_insert(maxproc, epset(m));
    std::vector< epset > upd_delete(maxproc, epset(m));

    std::vector< epset > subscriptions(maxproc, epset(n));
    std::vector< epset > updates(maxproc, epset(m));

    BEGIN_TIMER("Reduction");
#pragma omp parallel reduction(+:nmatches)
    {
        const int my_rank = omp_get_thread_num();
        const int thread_count = omp_get_num_threads();
        const size_t local_n = (n_endpoints + thread_count - 1) / thread_count;
        const size_t local_idx_start = my_rank*local_n;
        const size_t local_idx_end = std::min(local_idx_start + local_n, n_endpoints);

        // First step: local scans in parallel
#ifndef NDEBUG
#pragma omp critical
        std::cout << "Thread " << my_rank << " entering first block"
                  << " local_idx_start=" << local_idx_start
                  << " local_idx_end=" << local_idx_end
                  << std::endl;
#endif
        {
            epset& my_sub_insert( sub_insert[my_rank] );
            epset& my_sub_delete( sub_delete[my_rank] );
            epset& my_upd_insert( upd_insert[my_rank] );
            epset& my_upd_delete( upd_delete[my_rank] );

            for (size_t idx = local_idx_start; idx < local_idx_end; idx++) { 

                const endpoint& my_ep( endpoints[idx] );

                if (my_ep.t == endpoint::SUBSCRIPTION) {
                    if (my_ep.e == endpoint::LOWER) {
                        my_sub_insert.insert(my_ep.id);
                    } else {
                        if ( my_sub_insert.find(my_ep.id) )
                            my_sub_insert.remove(my_ep.id);
                        else 
                            my_sub_delete.insert(my_ep.id);
                    }
                } else {
                    if (my_ep.e == endpoint::LOWER) {
                        my_upd_insert.insert(my_ep.id);
                    } else {
                        if ( my_upd_insert.find(my_ep.id) )
                            my_upd_insert.remove(my_ep.id);
                        else
                            my_upd_delete.insert(my_ep.id);
                    }
                }
            } 
#ifndef NDEBUG
#pragma omp critical
            std::cout << "Thread " << my_rank << " exiting first block"
                      << " subs_insert=" << sub_insert[my_rank].size()
                      << " subs_delete=" << sub_delete[my_rank].size()
                      << std::endl;        
#endif
        }

#pragma omp barrier

        // Second step: global scan (master only)
#pragma omp master
        {
            for (int p=1; p<thread_count; p++) {
                subscriptions[p]
                    .merge( subscriptions[p-1] )
                    .merge( sub_insert[p-1] )
                    .subtract( sub_delete[p-1] );
                
                updates[p]
                    .merge( updates[p-1] )
                    .merge( upd_insert[p-1] )
                    .subtract( upd_delete[p-1] );

#ifndef NDEBUG
                std::cout << "p=" << p
                          << " subs=" << subscriptions[p].size()
                          << " upd=" << updates[p].size() << std::endl;
#endif
            }
        }

#pragma omp barrier

        // Third step: local scans (in parallel)
        {
            epset& my_sub = subscriptions[my_rank];
            epset& my_upd = updates[my_rank];

            for (size_t idx = local_idx_start; idx < local_idx_end; idx++) {
                const endpoint& my_ep = endpoints[idx];

                if (my_ep.t == endpoint::SUBSCRIPTION) {
                    if (my_ep.e == endpoint::LOWER) {
                        // std::cout << "S I " << my_ep.id << std::endl;
                        my_sub.insert(my_ep.id);
                    } else {
                        // std::cout << "S R " << my_ep.id << std::endl;
                        my_sub.remove(my_ep.id);
                        nmatches += my_upd.size();
                    }
                } else {
                    if (my_ep.e == endpoint::LOWER) {
                        // std::cout << "I R " << my_ep.id << std::endl;
                        my_upd.insert(my_ep.id);
                    } else {
                        // std::cout << "I R " << my_ep.id << std::endl;
                        my_upd.remove(my_ep.id);
                        nmatches += my_sub.size();
                    }
                } 
            }
            assert( my_sub.empty() );
            assert( my_upd.empty() );
        }
    }
    END_TIMER();

    return nmatches;
}

